//---------------------------------------------------------------------------
#include <vcl\vcl.h>
#pragma hdrstop

#include "TStructures_Editor.h"
#include "TDebuggingForm.h"
//---------------------------------------------------------------------------
#pragma resource "*.dfm"
TStructures_Editor *Structures_Editor;
//---------------------------------------------------------------------------
__fastcall TStructures_Editor::TStructures_Editor(TComponent* Owner)
    : TForm(Owner)
{
}
//---------------------------------------------------------------------------
void __fastcall TStructures_Editor::BitBtn1Click(TObject *Sender)
{
if (ImportFileDialog->Execute()) ImportFile( ImportFileDialog->FileName );
}
//---------------------------------------------------------------------------
void __fastcall TStructures_Editor::ImportFile( AnsiString pFileName )
{
WaitLabel->Caption = "Importing....";
WaitLabel->Refresh();
DebuggingForm->Caption = "Importing DirectX Structures.... debugging output";
DebuggingForm->ListBox->Items->Clear();

// import a standard-format text file
TStringList* strings = new TStringList();

// load the file into a temporary TStringList
// then copy it to another TStringList so that there are no blank rows
TStringList* temp_strings = new TStringList();
temp_strings->LoadFromFile( pFileName );
for (int i=0;i<temp_strings->Count;i++) if (!temp_strings->Strings[i].TrimLeft().TrimRight().IsEmpty()) strings->Add(temp_strings->Strings[i].TrimLeft().TrimRight());
delete temp_strings;

// convert it as required
AnsiString  file_type = "";
TStringList* structure_names = new TStringList();
TStringList* structure_descriptions = new TStringList();
TStringList* member_types = new TStringList();
TStringList* member_names = new TStringList();
TStringList* member_descriptions = new TStringList();
TStringList* flag_names = new TStringList();
TStringList* flag_descriptions = new TStringList();
AnsiString  temp_string = "";

// parse file :-
//
//  Structure
//  DDBLTBATCH
//  DDBLTFX
//  .
//  DDBLTBATCH
//  The DDBLTBATCH structure passes blit operations to the IDirectDrawSurface3::BltBatch method.
//  typedef struct _DDBLTBATCH{
//      LPRECT              lprDest;
//      LPDIRECTDRAWSURFACE lpDDSSrc;
//      LPRECT              lprSrc;
//      DWORD               dwFlags;
//      LPDDBLTFX           lpDDBltFx;
//  } DDBLTBATCH,FAR *LPDDBLTBATCH;
//  lprDest
//  Address of a RECT structure that defines the destination for the blit.
//  lpDDSSrc
//  Address of a DirectDrawSurface object that will be the source of the blit.
//  .
//  dwFlags
//  Optional control flags.
//  DDBLT_ALPHADEST
//  	Uses either the alpha information in pixel format or the alpha channel surface attached to the destination surface as the alpha channel for this blit.
//  DDBLT_ALPHADESTCONSTOVERRIDE
//  	Uses the dwAlphaDestConst member of the DDBLTFX structure as the alpha channel for the destination surface for this blit.
//  etc.
//  lpDDBltFx
//  Address of a DDBLTFX structure specifying additional blit effects.
//  .
//  QuickInfo
//
//    Windows NT: Use version 4.0 or later.
//    Windows: Use Windows 95 or later. Available as a redistributable for Windows 95.
//    Windows CE: Unsupported.
//    Header: Declared in ddraw.h.


// how we parse this file :-
// we scan through each of the 'strings' in the input file & parse as we go.
//

// check the file to make sure it contains structures
int row = 0;
if (strings->Strings[row] != "Structures")
    {
    DebuggingForm->ListBox->Items->Add( "Could not parse the import file '" + pFileName + "'.\n" +
                "You need to add 'Structures' or another type name to the top of the file before importing" );
    DebuggingForm->ListBox->Items->Add("ERROR LINE: " + strings->Strings[row] );
    DebuggingForm->Show();
    WaitLabel->Caption = "";
    return;
    }
row++;
//DebuggingForm->ListBox->Items->Add( "Found the keyword 'Structures' successfully");

// Immediately following the 'Structures' keyword is a list of the different
// structures defined inside the file.
// By scanning rows for structure names we build a list of the various
// structures that we are going to parse.
// When a structure name is repeated it signifies that the first structure
// definition is following
bool found_first_definition = false;
do
    {
    //DebuggingForm->ListBox->Items->Add( "Parsing structure name '" + strings->Strings[row] + "'" );
    // check the current row to see if it contains a unique structure name
    bool is_already_structure_name = false;
    for (int i=0;i<structure_names->Count;i++)
        if (strings->Strings[row]==structure_names->Strings[i])
            {
            //DebuggingForm->ListBox->Items->Add( "Structure name '" + strings->Strings[row] + "' already exists" );
            is_already_structure_name = true;
            break;  // continue at end of this for loop
            }
    // if we've already added the name previously
    if (is_already_structure_name)
        {
        //DebuggingForm->ListBox->Items->Add( "Assuming structure definition follows" );
        found_first_definition = true;
        break;      // continue at end of while loop
        }
    //DebuggingForm->ListBox->Items->Add( "Adding '" + strings->Strings[row] + "' to structure names that are to be parsed" );
    DebuggingForm->ListBox->Items->Add("STRUCTURE NAME: " + strings->Strings[row]);
    structure_names->Add( strings->Strings[row] );
    row++;
    if (row == strings->Count)
        {
        DebuggingForm->ListBox->Items->Add( "Could not parse the import file '" + pFileName + "'.\n" +
                                "There is something wrong with the structure header block (top of file)\n" +
                                "There is nothing to denote a definition follows." );
        }
    }
while (!found_first_definition);

// since we now know how many structures we need to parse, attempt to parse that many structures
for (int structure=0;structure<structure_names->Count;structure++)
    {
    WaitLabel->Caption = "Importing.... " + structure_names->Strings[structure];
    WaitLabel->Refresh();
    DebuggingForm->ListBox->Items->Add( ">>>>>> Parsing definition of structure '" + structure_names->Strings[structure] + "' <<<<<<" );

//    DebuggingForm->ListBox->Items->Add( "Clearing structure internal representations" );
    member_types->Clear();
    member_names->Clear();
    member_descriptions->Clear();
    flag_names->Clear();
    flag_descriptions->Clear();

    // since we already know the names, check that the definition is appropriate but otherwise
    // ignore this row.
    bool structure_is_appropriate = false;
    for (int i=0;i<structure_names->Count;i++)
        if (structure_names->Strings[i]==strings->Strings[row])
            {
            structure_is_appropriate = true;
            break;
            }
    if (!structure_is_appropriate)
        {
        DebuggingForm->ListBox->Items->Add("Could not parse the import file '" + pFileName + "'.\n" +
                    "There is a structure '" + strings->Strings[row] + "'\n" +
                    "being defined but there has been no previous definition at the start of the file." );
        DebuggingForm->ListBox->Items->Add("ERROR LINE: " + strings->Strings[row] );
        DebuggingForm->Show();
        WaitLabel->Caption = "";
        return;
        }
//    DebuggingForm->ListBox->Items->Add( "Structure name tallies with a definition in the structure header block");
    row++;

    // simply add the structure description
//    DebuggingForm->ListBox->Items->Add( "Adding the description '" + strings->Strings[row] + "'" );
    structure_descriptions->Add( strings->Strings[row] );
    row++;

//    DebuggingForm->ListBox->Items->Add( "STRUCTURE NAME: " + structure_names->Strings[structure] );
    DebuggingForm->ListBox->Items->Add( "STRUCTURE [" + structure_names->Strings[structure] + "] DESCRIPTION: " + structure_descriptions->Strings[structure] );
    
    // by all rights we _must_ have a "struct" somewhere in this row
//    DebuggingForm->ListBox->Items->Add( "Checking row for a 'struct' construct" );
//    DebuggingForm->ListBox->Items->Add( "Row is ::" +strings->Strings[row]+"::" );
    if (strings->Strings[row].Pos("struct")==0)
        {
        DebuggingForm->ListBox->Items->Add("Could not parse the import file '" + pFileName + "'.\n" +
                    "There is no starting point for the 'struct' definition for the object '" + structure_names->Strings[structure] + "'" );
        DebuggingForm->ListBox->Items->Add("ERROR LINE: " + strings->Strings[row] );
        DebuggingForm->Show();
        WaitLabel->Caption = "";
        return;
        }
//    DebuggingForm->ListBox->Items->Add( "A 'struct' definition has been found.... '" + strings->Strings[row] + "'"  );
    row++;

    // the following rows may occur anywhere.... hence we loop till we find the end of the 'struct' definition
    bool end_of_struct = false;
//    DebuggingForm->ListBox->Items->Add( "Parsing 'struct' as follows :- " );
    do
        {
        // skip out when we find the end-token
        if (strings->Strings[row].Pos("} ")!=0)
            {
//            DebuggingForm->ListBox->Items->Add( "Found the end of the structure." );
            row++;
            end_of_struct = true;
            continue;
            }

        // skip anything to do with a 'union' declaration
        if (strings->Strings[row].Pos("union")!=0)
            {
//            DebuggingForm->ListBox->Items->Add( "Found the start of a 'union' declaration" );
            row++;     // ignore any 'union' declarations
            continue;
            }
        if (strings->Strings[row].Pos("{")!=0)
            {
//            DebuggingForm->ListBox->Items->Add( "Found the start bracket for a 'union' declaration" );
            row++;         // ignore any '{' declarations.. part of the 'union'
            continue;
            }
        if (strings->Strings[row].Pos("};")!=0)
            {
//            DebuggingForm->ListBox->Items->Add( "Found the end bracket for a 'union' declaration" );
            row++;        // ignore any '};' declarations  part of the 'union'
            continue;
            }

        // parse members of the structure
        // (1)
        //     LPRECT              lprDest;         // some comment
        // LPRECT              lprDest;         // some comment
        temp_string = strings->Strings[row].TrimLeft();
        // (2)
        // LPRECT              lprDest;         // some comment
        //       ^
        int temp_pos = temp_string.Pos(" ");
        // LPRECT
        AnsiString member_type = temp_string.SubString(1,temp_pos-1);
//        DebuggingForm->ListBox->Items->Add( "temp_string == '" + temp_string + "'" );
//        DebuggingForm->ListBox->Items->Add( "temp_pos == " + IntToStr(temp_pos));
//        DebuggingForm->ListBox->Items->Add( "member_type == '" + member_type + "'" );
        DebuggingForm->ListBox->Items->Add( "STRUCTURE [" + structure_names->Strings[structure] + "] MEMBER TYPE: '"  + member_type + "'" );
        member_types->Add( member_type );
        // lprDest;         // some comment
        AnsiString member_name = temp_string.SubString(temp_pos, temp_string.Length() - temp_pos ).TrimLeft();
//        DebuggingForm->ListBox->Items->Add( "member_name == '" + member_name + "'" );
        temp_pos = member_name.Pos(";");
        if (temp_pos!=0) member_name = member_name.SubString( 1, temp_pos-1 );
//        DebuggingForm->ListBox->Items->Add( "member_name - semicolon == '" + member_name + "'" );
        // lprDest;         // some comment
        //        ^
        // lprDest
        DebuggingForm->ListBox->Items->Add( "STRUCTURE [" + structure_names->Strings[structure] + "] MEMBER NAME: '"+ member_name + "'" );
        member_names->Add( member_name );

        row++;
        if (row >= strings->Count)
            {
            DebuggingForm->ListBox->Items->Add("Could not parse the import file '" + pFileName + "'.\n" +
                        "Could not find the end of a 'struct' definition." );
            DebuggingForm->ListBox->Items->Add("ERROR LINE: " + strings->Strings[row] );
            DebuggingForm->Show();
            WaitLabel->Caption = "";
            return;
            }
        }
    while (!end_of_struct);

    // we now expect the word "Members"... if not something has gone wrong.
    if (strings->Strings[row]!="Members")
        {
        DebuggingForm->ListBox->Items->Add("Could not parse the import file '" + pFileName + "'.\n" +
                    "Could not find the 'Members' keyword." );
        DebuggingForm->ListBox->Items->Add("ERROR LINE: " + strings->Strings[row] );
        DebuggingForm->Show();
        WaitLabel->Caption = "";
        return;
        }
    row++;
//    DebuggingForm->ListBox->Items->Add("Found the keyword 'Members'");

    // we can now parse the descriptions of the various members of the structure.
    // note that some members also contain descriptions of various flags.
    // note also that other members contain "something and somethingelse" which
    // we need to distinguish between.
//    DebuggingForm->ListBox->Items->Add("Attempting to parse the various " + IntToStr( member_names->Count ) + " 'Members'" );
    bool checking_last_member_flags = false;
    for (int member=0;member<member_names->Count;member++)
        {
        DebuggingForm->ListBox->Items->Add("Parsing the " + member_names->Strings[member] + " Member" );
        // check for the premature end of the members descriptions
        if ((strings->Strings[row]=="QuickInfo") ||
            (strings->Strings[row]=="Remarks"))
            {
            // special case: if we're checking for flags for the last member, and the check failed
            // do not check for "this and that" rules.
            if (checking_last_member_flags)
                {
                DebuggingForm->ListBox->Items->Add("Found a 'QuickInfo' or 'Remarks' section after checking last member flags.");
                DebuggingForm->ListBox->Items->Add("Assuming structure is finished....");
                break;
                }

            // there's an error because we haven't yet parsed all of the relevant members
            DebuggingForm->ListBox->Items->Add("Could not parse the import file '" + pFileName + "'.\n" +
                                "Found the 'QuickInfo' keyword before parsing the various member descriptions....");
            DebuggingForm->ListBox->Items->Add("ERROR LINE: " + strings->Strings[row] );
            DebuggingForm->Show();
            WaitLabel->Caption = "";
            return;
            }

        // if the current line contains the name of the appropriate member, ignore the member name &
        // store the description that follows immediately after this row.
        if (strings->Strings[row]==member_names->Strings[member])
            {
            row++;
//            DebuggingForm->ListBox->Items->Add("Found a valid member name: '" + member_names->Strings[member] + "'. Adding description '" + strings->Strings[row] + "'" );
            member_descriptions->Add( strings->Strings[row] );
            row++;
            DebuggingForm->ListBox->Items->Add("STRUCTURE [" + structure_names->Strings[structure] + "] MEMBER [" + member_names->Strings[member] + "] DESCRIPTION: " + member_descriptions->Strings[member] );

            // special case.... for the last member, it may still have flags... must check for flags also.
            // simplest way of doing this is to redo the loop (once).
            if (member==member_names->Count-1)
                {
                checking_last_member_flags = true;
                member--;
                }
            }
        else
            // it's not a simple member....
            // maybe it's a FLAGS parameter.
            // how do we check? all FLAGS are declared in uppercase.
            // if not, maybe it's a case of "this and that"....
            {
            DebuggingForm->ListBox->Items->Add( "Found a non-valid member name where a member should be... " );
            DebuggingForm->ListBox->Items->Add( "Original string is :- '" + strings->Strings[row] + "'" );
            DebuggingForm->ListBox->Items->Add( "Member name is :- '" + member_names->Strings[member] + "'" );
            DebuggingForm->ListBox->Items->Add( "Checking FLAGS rules" );
            // case-sensive comparison of normal & UpperCase() versions. 0 == exactly the same
            if (strings->Strings[row].AnsiCompare( UpperCase(strings->Strings[row])) == 0)
                {
                DebuggingForm->ListBox->Items->Add( "The unknown member name fits the rule for a FLAGS description.. Temporarily reverting to previous member" );
                if (member>0) member--;
                DebuggingForm->ListBox->Items->Add( "Adding flags for member '" + member_names->Strings[member] + "'" );
                do
                    {
                    // it's probably a FLAGS description....
                    DebuggingForm->ListBox->Items->Add("STRUCTURE [" + structure_names->Strings[structure] + "] MEMBER [" + member_names->Strings[member] + "] FLAG NAME: " + strings->Strings[row] );
                    // add a gnarled version of the flag name that contains the structure & member index....
                    flag_names->Add( IntToStr(structure) + "," + IntToStr(member) + "::" + strings->Strings[row] );
                    row++;
                    DebuggingForm->ListBox->Items->Add("STRUCTURE [" + structure_names->Strings[structure] + "] MEMBER [" + member_names->Strings[member] + "] FLAG DESCRIPTION: '" + strings->Strings[row] );
                    flag_descriptions->Add( strings->Strings[row] );
                    row++;
                    }
                while (strings->Strings[row].AnsiCompare( UpperCase(strings->Strings[row])) == 0);
                }
            else
                // maybe it's a "this and that"? or a "this, that and something_else" format
                {
                DebuggingForm->ListBox->Items->Add("FLAGS rules did not apply.");
                DebuggingForm->ListBox->Items->Add("Checking 'this and that' rules");
                // if there is another member name following this one....
                if (member<member_names->Count-1)
                    {
                    // determine if the string is in a "this and that" format...
                    int temp_pos = strings->Strings[row].Pos(" and ");
                    if ( temp_pos != 0 )
                        {
                        AnsiString first_string = strings->Strings[row].SubString(1,temp_pos-1);
                        AnsiString second_string = strings->Strings[row].SubString( temp_pos+5, strings->Strings[row].Length()-temp_pos+1 );
                        DebuggingForm->ListBox->Items->Add("First String is :- '" + first_string + "'" );
                        DebuggingForm->ListBox->Items->Add("Second String is :- '" + second_string + "'" );
                        // determine the "that"
                        DebuggingForm->ListBox->Items->Add("Current member name is :- '" + member_names->Strings[member] + "'" );
                        DebuggingForm->ListBox->Items->Add("Next member name is :- '" + member_names->Strings[member+1] + "'" );

                        // check if it's a "this and that" format
                        if ((first_string == member_names->Strings[member]) && (second_string == member_names->Strings[member+1]))
                            {
                            DebuggingForm->ListBox->Items->Add( "Successfully found a match. '" + first_string + "' and '" + second_string + "'" );
                            DebuggingForm->ListBox->Items->Add("Adding member_descriptions for both '" + first_string + " and " + second_string + "'" );
                            DebuggingForm->ListBox->Items->Add("");
                            row++;
                            member_descriptions->Add( strings->Strings[row] );
                            member_descriptions->Add( strings->Strings[row] );
                            member++;
                            row++;
                            }
                        else
                            // maybe it's a "this, that and something_else" format?
                            {
                            DebuggingForm->ListBox->Items->Add("Didn't match the 'this and that' rule.");
                            DebuggingForm->ListBox->Items->Add("maybe it fits the 'this, that and something_else' rule");
                            // if there are at least 2 more member names following this one....
                            if (member<member_names->Count-2)
                                {
                                // check if it's a "this, that and something_else" format
                                temp_pos = first_string.Pos(",");
                                if (temp_pos!=0)
                                    {
                                    AnsiString leftofcomma = first_string.SubString(1,temp_pos-1);
                                    AnsiString rightofcomma = first_string.SubString(temp_pos+1,first_string.Length()-temp_pos-1).TrimLeft();
                                    DebuggingForm->ListBox->Items->Add("first_string, leftofcomma = '" + leftofcomma + "'" );
                                    DebuggingForm->ListBox->Items->Add("first_string, rightofcomma = '" + rightofcomma + "'" );
                                    DebuggingForm->ListBox->Items->Add("second_string = '" + second_string + "'" );
                                    DebuggingForm->ListBox->Items->Add("current member = '" + member_names->Strings[member] + "'" );
                                    DebuggingForm->ListBox->Items->Add("next member = '" + member_names->Strings[member+1] + "'" );
                                    DebuggingForm->ListBox->Items->Add("second member = '" + member_names->Strings[member+2] + "'" );
                                    if (leftofcomma == member_names->Strings[member]) DebuggingForm->ListBox->Items->Add("first param '" + leftofcomma + "'is the same");
                                    if (rightofcomma == member_names->Strings[member+1]) DebuggingForm->ListBox->Items->Add("second param '" + rightofcomma + "'is the same");
                                    if (second_string == member_names->Strings[member+2]) DebuggingForm->ListBox->Items->Add("third param '" + second_string + "'is the same");
                                    if ((leftofcomma == member_names->Strings[member]) &&
                                        (rightofcomma == member_names->Strings[member+1]) &&
                                        (second_string == member_names->Strings[member+2]))
                                        {
                                        DebuggingForm->ListBox->Items->Add("Found a 'this, that and something_else' format");
                                        DebuggingForm->ListBox->Items->Add("Adding member_descriptions for '" + leftofcomma + ", " + rightofcomma + " and " + second_string + "'" );
                                        DebuggingForm->ListBox->Items->Add("");
                                        row++;
                                        member_descriptions->Add( strings->Strings[row] );
                                        member_descriptions->Add( strings->Strings[row] );
                                        member_descriptions->Add( strings->Strings[row] );
                                        member++;
                                        member++;
                                        row++;
                                        }
                                    else
                                        {
                                        DebuggingForm->ListBox->Items->Add("Could not parse the import file '" + pFileName + "'.\n" +
                                                        "Was not able to figure out the 'this, that and something_else' rules\n" +
                                                        "Line: " + strings->Strings[row] );
                                        DebuggingForm->ListBox->Items->Add("ERROR LINE: " + strings->Strings[row] );
                                        DebuggingForm->Show();
                                        WaitLabel->Caption = "";
                                        return;
                                        }
                                    }
                                else
                                    {
                                    DebuggingForm->ListBox->Items->Add("Could not parse the import file '" + pFileName + "'.\n" +
                                                    "Was not able to figure out the rules\n" +
                                                    "Line: " + strings->Strings[row] );
                                    DebuggingForm->ListBox->Items->Add("ERROR LINE: " + strings->Strings[row] );
                                    DebuggingForm->Show();
                                    WaitLabel->Caption = "";
                                    return;
                                    }
                                }
                            else
                                {
                                DebuggingForm->ListBox->Items->Add("Could not parse the import file '" + pFileName + "'.\n" +
                                                "Was not able to figure out the latter half of a 'this & that' rule\n" +
                                                "Line: " + strings->Strings[row] );
                                DebuggingForm->ListBox->Items->Add("ERROR LINE: " + strings->Strings[row] );
                                DebuggingForm->Show();
                                WaitLabel->Caption = "";
                                return;
                                }
                            }
                        }
                    else
                        {
                        DebuggingForm->ListBox->Items->Add("Could not parse the import file '" + pFileName + "'.\n" +
                                        "Could not find a valid 'this and that' rule :- \n" +
                                        strings->Strings[row] );
                        DebuggingForm->ListBox->Items->Add("ERROR LINE: " + strings->Strings[row] );
                        DebuggingForm->Show();
                        WaitLabel->Caption = "";
                        return;
                        }
                    }
                else
                    {
                    DebuggingForm->ListBox->Items->Add("Could not parse the import file '" + pFileName + "'.\n" +
                                "Could not interpret a line using any known rules\n" );
                    DebuggingForm->ListBox->Items->Add("ERROR LINE: " + strings->Strings[row] );
                    DebuggingForm->Show();
                    WaitLabel->Caption = "";
                    return;
                    }
                }
            }
        }
    // we now expect either a "Remarks" or a "QuickInfo" section
    // check just to make sure

    // if it is a 'Remarks' section, simply skip it till we find a "QuickInfo" section.
    if (strings->Strings[row]=="Remarks")
        while (strings->Strings[row]!="QuickInfo")
            {
            row++;
            if (row>=strings->Count)
                {
                DebuggingForm->ListBox->Items->Add("Could not parse the import file '" + pFileName + "'.\n" +
                               "Found a 'Remarks' section but did not find the 'QuickInfo' section following" +
                                strings->Strings[row] );
                DebuggingForm->ListBox->Items->Add("ERROR LINE: " + strings->Strings[row] );
                DebuggingForm->Show();
                WaitLabel->Caption = "";
                return;
                }
            }

//    DebuggingForm->ListBox->Items->Add("Checking if the keyword 'QuickInfo' exists.");
//    DebuggingForm->ListBox->Items->Add("string being checked == '" + strings->Strings[row] + "'" );
    if (strings->Strings[row]!="QuickInfo")
        {
        DebuggingForm->ListBox->Items->Add("Could not parse the import file '" + pFileName + "'.\n" +
                    "Could not find a 'QuickInfo' section" );
        DebuggingForm->ListBox->Items->Add("ERROR LINE: " + strings->Strings[row] );
        DebuggingForm->Show();
        WaitLabel->Caption = "";
        return;
        }

    // skip past the four lines
//    DebuggingForm->ListBox->Items->Add("The keyword 'QuickInfo' exists. skipping the 'QuickInfo' block");
    row += 5;

    // yey. the structure definition has been successfully parsed.
    DebuggingForm->ListBox->Items->Add( ">>>>>> Definition of structure '" + structure_names->Strings[structure] + "' was completed successfully! <<<<<<" );

    // since it's all been successful.
    DebuggingForm->ListBox->Items->Add("Saving structure definition for '" + structure_names->Strings[structure] + "' into the database....");
    Structures_Table->Close();       // close and re-open so that we know for sure that the table has been opened
    Structures_Table->Open();
    Structures_Table->Append();
    Structures_Table->FieldByName("Conversion_Status")->AsInteger = 0;
    Structures_Table->FieldByName("Interface_Index")->AsInteger = fInterface_Index;
    Structures_Table->FieldByName("Structure_Name")->AsString = structure_names->Strings[structure];
    Structures_Table->FieldByName("Structure_Description")->AsString = structure_descriptions->Strings[structure];
    Structures_Table->FieldByName("BCB_Name")->AsString = "T" + structure_names->Strings[structure].LowerCase();
    Structures_Table->FieldByName("BCB_Description")->AsString = structure_descriptions->Strings[structure];
    Structures_Table->Post();

    TempQuery->Close();
    TempQuery->SQL->Clear();
    TempQuery->SQL->Add(" select Structure_Index from Structures.DB" );
    TempQuery->SQL->Add(" where Structure_Name = '" + structure_names->Strings[ structure ] + "'" );
    TempQuery->Open();
    int structure_index = TempQuery->FieldByName("Structure_Index")->AsInteger;
    TempQuery->Close();

    StructureMembers_Table->Close();
    StructureMembers_Table->Open();
    for (int i=0;i<member_names->Count;i++)
        {
        StructureMembers_Table->Append();
        DebuggingForm->ListBox->Items->Add("Saving structure member '" + structure_names->Strings[structure] + "." + member_names->Strings[i] + "'" );
        StructureMembers_Table->FieldByName("Interface_Index")->AsInteger = fInterface_Index;
        StructureMembers_Table->FieldByName("Structure_Index")->AsInteger = structure_index;
        StructureMembers_Table->FieldByName("Member_Name")->AsString = member_names->Strings[i];
        StructureMembers_Table->FieldByName("Member_Type")->AsString = member_types->Strings[i];
        StructureMembers_Table->FieldByName("Member_Description")->AsString= member_descriptions->Strings[i];
        // change the following line to put bcb-default name....
        // ignore chars till first non-lowercase character of member_names->Strings[i]
        StructureMembers_Table->FieldByName("BCB_Name")->AsString = member_names->Strings[i];
        StructureMembers_Table->FieldByName("BCB_Description")->AsString= member_descriptions->Strings[i];
        StructureMembers_Table->Post();
        }

    StructureMemberFlags_Table->Close();
    StructureMemberFlags_Table->Open();
    for (int j=0;j<flag_names->Count;j++)
        {
        StructureMemberFlags_Table->Append();
        int temp_pos = flag_names->Strings[j].Pos("::");
        AnsiString indexes = flag_names->Strings[j].SubString( 1, temp_pos-1 );
        AnsiString flag_name = flag_names->Strings[j].SubString( temp_pos+2, flag_names->Strings[j].Length() - temp_pos - 1 );
        temp_pos = indexes.Pos(",");
        AnsiString relevant_structure_index = indexes.SubString( 1, temp_pos-1 );
        AnsiString relevant_member_index = indexes.SubString( temp_pos+1, indexes.Length() - temp_pos );

        DebuggingForm->ListBox->Items->Add("Saving structure member flag '" + structure_names->Strings[structure] + "." + member_names->Strings[ StrToInt(relevant_member_index) ] + ":" + flag_name + "'" );

        TempQuery->Close();
        TempQuery->SQL->Clear();
        TempQuery->SQL->Add(" select Member_Index from Structure_Members.DB" );
        TempQuery->SQL->Add(" where Member_Name = '" + member_names->Strings[ StrToInt(relevant_member_index) ] + "'" );
        TempQuery->SQL->Add(" and Structure_Index = " + IntToStr(structure_index) );
        TempQuery->Open();
        int member_index = TempQuery->FieldByName("Member_Index")->AsInteger;
        TempQuery->Close();

        StructureMemberFlags_Table->FieldByName("Interface_Index")->AsInteger = fInterface_Index;
        StructureMemberFlags_Table->FieldByName("Structure_Index")->AsInteger = structure_index;
        StructureMemberFlags_Table->FieldByName("Member_Index")->AsInteger = member_index;
        StructureMemberFlags_Table->FieldByName("Flag_Name")->AsString = flag_name;
        StructureMemberFlags_Table->FieldByName("Flag_Description")->AsString = flag_descriptions->Strings[j];
        StructureMemberFlags_Table->FieldByName("BCB_Description")->AsString = flag_descriptions->Strings[j];
        StructureMemberFlags_Table->Post();
        }
    DebuggingForm->ListBox->Items->Add("NOTE: Structure definition for '" + structure_names->Strings[structure] + "' has been saved successfully" );

    //    DebuggingForm->ListBox->Items->Add( "Clearing structure internal representations" );
    member_types->Clear();
    member_names->Clear();
    member_descriptions->Clear();
    flag_names->Clear();
    flag_descriptions->Clear();
    }

DebuggingForm->ListBox->Items->Add( "___ ALL FINISHED ___" );
ShowMessage("Structures were imported SUCCESSFULLY!");
WaitLabel->Caption = "";
DebuggingForm->Show();

// clean up usage
delete flag_descriptions;
delete flag_names;
delete member_descriptions;
delete member_names;
delete member_types;
delete structure_descriptions;
delete structure_names;
delete strings;
}
//---------------------------------------------------------------------------
void __fastcall TStructures_Editor::BuildSelections()
{
BuildStructureSelections();
BuildMemberSelections();
BuildFlagSelections();
}
//---------------------------------------------------------------------------
void __fastcall TStructures_Editor::BuildStructureSelections()
{
// build a list of current known structures & put them in the DirectX_ComboBox & BCB_ComboBox
TempQuery->Close();
TempQuery->SQL->Clear();
TempQuery->SQL->Add(" select Structure_Name, BCB_Name from Structures.DB s, Interfaces.DB i " );
TempQuery->SQL->Add(" where s.Interface_Index = i.Interface_Index " );
TempQuery->SQL->Add(" and i.Interface_Name = '" + Label1->Caption + "'" );
TempQuery->SQL->Add(" order by Structure_Name ");
TempQuery->Open();
TempQuery->First();
DirectX_ComboBox->Items->Clear();
BCB_ComboBox->Items->Clear();
while (!TempQuery->Eof)
    {
    DirectX_ComboBox->Items->Add(TempQuery->FieldByName("Structure_Name")->AsString);
    BCB_ComboBox->Items->Add(TempQuery->FieldByName("BCB_Name")->AsString);
    TempQuery->Next();
    }
TempQuery->Close();
DirectX_ComboBox->ItemIndex = 0;
BCB_ComboBox->ItemIndex = 0;
}
//---------------------------------------------------------------------------
void __fastcall TStructures_Editor::BuildMemberSelections()
{
// fill in the list of known structure members & put them in the DirectX_StructureMember_ComboBox & BCB_ObjectProperty_ComboBox
TempQuery->Close();
TempQuery->SQL->Clear();
TempQuery->SQL->Add(" select Member_Name, BCB_Name from Structure_Members.DB" );
TempQuery->SQL->Add(" where Structure_Index = " + IntToStr( fStructure_Index ) );
TempQuery->SQL->Add(" order by Member_Name ");
TempQuery->Open();
TempQuery->First();
DirectX_StructureMember_ComboBox->Items->Clear();
BCB_ObjectProperty_ComboBox->Items->Clear();
while (!TempQuery->Eof)
    {
    DirectX_StructureMember_ComboBox->Items->Add(TempQuery->FieldByName("Member_Name")->AsString);
    BCB_ObjectProperty_ComboBox->Items->Add(TempQuery->FieldByName("BCB_Name")->AsString);
    TempQuery->Next();
    }
TempQuery->Close();
DirectX_StructureMember_ComboBox->ItemIndex = 0;
BCB_ObjectProperty_ComboBox->ItemIndex = 0;
}
//---------------------------------------------------------------------------
void __fastcall TStructures_Editor::BuildFlagSelections()
{
// fill in the list of known member flags & put them in the DirectX_MemberFlags_ComboBox & BCB_PropertyFlags_ComboBox
TempQuery->Close();
TempQuery->SQL->Clear();
TempQuery->SQL->Add(" select Flag_Name from Structure_Member_Flags.DB " );
TempQuery->SQL->Add(" where Member_Index = " + IntToStr( fMember_Index ) );
TempQuery->SQL->Add(" order by Flag_Name ");
TempQuery->Open();
TempQuery->First();
DirectX_MemberFlags_ComboBox->Items->Clear();
BCB_PropertyFlags_ComboBox->Items->Clear();
while (!TempQuery->Eof)
    {
    DirectX_MemberFlags_ComboBox->Items->Add(TempQuery->FieldByName("Flag_Name")->AsString);
    BCB_PropertyFlags_ComboBox->Items->Add(TempQuery->FieldByName("Flag_Name")->AsString);
    TempQuery->Next();
    }
TempQuery->Close();
DirectX_MemberFlags_ComboBox->ItemIndex = 0;
BCB_PropertyFlags_ComboBox->ItemIndex = 0;
DirectX_MemberFlags_GroupBox->Visible = (DirectX_MemberFlags_ComboBox->Items->Count!=0);
BCB_ObjectFlags_GroupBox->Visible = (DirectX_MemberFlags_ComboBox->Items->Count!=0);
}
//---------------------------------------------------------------------------
void __fastcall TStructures_Editor::FormActivate(TObject *Sender)
{
DirectX_To_BCB->Connected = true;
Structures_Table->Active = true;
StructureMembers_Table->Active = true;
StructureMemberFlags_Table->Active = true;
if (Structures_Table->RecordCount==0) Structures_Table->Insert(); else Structures_Table->Edit();
if (StructureMembers_Table->RecordCount==0) StructureMembers_Table->Insert(); else StructureMembers_Table->Edit();
if (StructureMemberFlags_Table->RecordCount==0) StructureMemberFlags_Table->Insert(); else StructureMemberFlags_Table->Edit();

// get the relevant Interfaces_Index to be inserted into the relevant files.
TempQuery->Close();
TempQuery->SQL->Clear();
TempQuery->SQL->Add( " select Interface_Index from Interfaces.DB" );
TempQuery->SQL->Add( " where Interface_Name = '" + Label1->Caption + "'" );
TempQuery->Open();
fInterface_Index = TempQuery->FieldByName("Interface_Index")->AsInteger;
TempQuery->Close();

BuildStructureSelections();
DirectX_ComboBox->ItemIndex = 0;
DirectX_ComboBoxChange(Sender);
}
//---------------------------------------------------------------------------
void __fastcall TStructures_Editor::FormDeactivate(TObject *Sender)
{
Structures_Table->Active = false;
StructureMembers_Table->Active = false;
StructureMemberFlags_Table->Active = false;
DirectX_To_BCB->Connected = false;
}
//---------------------------------------------------------------------------
void __fastcall TStructures_Editor::FormCreate(TObject *Sender)
{
Top = 0;
Left = 0;
Width = 800;
Height = 600;
}
//---------------------------------------------------------------------------
void __fastcall TStructures_Editor::DirectX_ComboBoxChange(TObject *Sender)
{
if (DirectX_ComboBox->ItemIndex!=-1)
    {
    BCB_ComboBox->ItemIndex = DirectX_ComboBox->ItemIndex;
    // note: seems a bit wierd to use this query here.... but ....
    // even though the Structure_Name field is declared as an index (using Database Desktop) &
    // the Structure_Name field is actually determined as an index by the object inspector, when we
    // attempt to use it as a search key for GotoKey(), etc. BCB responds
    // "there is no key field called Structure_Name" ... huh?
    // hence we retrieve an use a valid index field manually.

    TempQuery->Close();
    TempQuery->SQL->Clear();
    TempQuery->SQL->Add(" select Structure_Index from Structures.DB " );
    TempQuery->SQL->Add(" where Structure_Name = '" + DirectX_ComboBox->Items->Strings[DirectX_ComboBox->ItemIndex] + "'" );
    TempQuery->Open();
    fStructure_Index = TempQuery->FieldByName("Structure_Index")->AsInteger;

    TempQuery->Close();

    Structures_Table->Close();
    Structures_Table->IndexFieldNames = "Structure_Index";
    Structures_Table->Open ();

    Structures_Table->SetKey();
    Structures_Table->FieldByName("Structure_Index")->AsInteger = fStructure_Index;
    Structures_Table->GotoNearest();

    BuildMemberSelections();
    BuildFlagSelections();
    DirectX_StructureMember_ComboBoxChange(Sender);
    DirectX_MemberFlags_ComboBoxChange(Sender);
    }
}
//---------------------------------------------------------------------------
void __fastcall TStructures_Editor::DirectX_StructureMember_ComboBoxChange(
    TObject *Sender)
{
if ((DirectX_ComboBox->ItemIndex!=-1) &&
    (DirectX_StructureMember_ComboBox->ItemIndex!=-1))
    {
    BCB_ObjectProperty_ComboBox->ItemIndex = DirectX_StructureMember_ComboBox->ItemIndex;

    // note: seems a bit wierd to use this query here.... but ....
    // even though the Structure_Name field is declared as an index (using Database Desktop) &
    // the Structure_Name field is actually determined as an index by the object inspector, when we
    // attempt to use it as a search key for GotoKey(), etc. BCB responds
    // "there is no key field called Structure_Name" ... huh?
    // hence we retrieve an use a valid index field manually.

    TempQuery->Close();
    TempQuery->SQL->Clear();
    TempQuery->SQL->Add(" select Member_Index from Structure_Members.DB " );
    TempQuery->SQL->Add(" where Member_Name = '" + DirectX_StructureMember_ComboBox->Items->Strings[DirectX_StructureMember_ComboBox->ItemIndex] + "'" );
    TempQuery->Open();
    fMember_Index = TempQuery->FieldByName("Member_Index")->AsInteger;

    TempQuery->Close();

    StructureMembers_Table->Close();
    StructureMembers_Table->IndexFieldNames = "Member_Index";
    StructureMembers_Table->Open ();

    StructureMembers_Table->SetKey();
    StructureMembers_Table->FieldByName("Member_Index")->AsInteger = fMember_Index;
    StructureMembers_Table->GotoNearest();

    BuildFlagSelections();
    }
}
//---------------------------------------------------------------------------
void __fastcall TStructures_Editor::DirectX_MemberFlags_ComboBoxChange(
    TObject *Sender)
{
if ((DirectX_ComboBox->ItemIndex!=-1) &&
    (DirectX_StructureMember_ComboBox->ItemIndex!=-1) &&
    (DirectX_MemberFlags_ComboBox->ItemIndex!=-1))
    {
    BCB_PropertyFlags_ComboBox->ItemIndex = DirectX_MemberFlags_ComboBox->ItemIndex;
    // note: seems a bit wierd to use this query here.... but ....
    // even though the Structure_Name field is declared as an index (using Database Desktop) &
    // the Structure_Name field is actually determined as an index by the object inspector, when we
    // attempt to use it as a search key for GotoKey(), etc. BCB responds
    // "there is no key field called Structure_Name" ... huh?
    // hence we retrieve an use a valid index field manually.

    TempQuery->Close();
    TempQuery->SQL->Clear();
    TempQuery->SQL->Add(" select Flag_Index from Structure_Member_Flags.DB " );
    TempQuery->SQL->Add(" where Flag_Name = '" + DirectX_MemberFlags_ComboBox->Items->Strings[DirectX_MemberFlags_ComboBox->ItemIndex] + "'" );
    TempQuery->Open();
    int flag_index = TempQuery->FieldByName("Flag_Index")->AsInteger;

    TempQuery->Close();

    StructureMemberFlags_Table->Close();
    StructureMemberFlags_Table->IndexFieldNames = "Flag_Index";
    StructureMemberFlags_Table->Open ();

    StructureMemberFlags_Table->SetKey();
    StructureMemberFlags_Table->FieldByName("Flag_Index")->AsInteger = flag_index;
    StructureMemberFlags_Table->GotoNearest();
    }
}
//---------------------------------------------------------------------------
void __fastcall TStructures_Editor::BitBtn3Click(TObject *Sender)
{
ShowMessage("NOTE: Clearing _ALL_ Structures, Methods, Flags, etc.");
Structures_Table->Close();
Structures_Table->EmptyTable();
Structures_Table->Open();
StructureMembers_Table->Close();
StructureMembers_Table->EmptyTable();
StructureMembers_Table->Open();
StructureMemberFlags_Table->Close();
StructureMemberFlags_Table->EmptyTable();
StructureMemberFlags_Table->Open();

BuildSelections();
}
//---------------------------------------------------------------------------
void __fastcall TStructures_Editor::BitBtn6Click(TObject *Sender)
{
if (DirectX_ComboBox->ItemIndex > 0)
    {
    DirectX_ComboBox->ItemIndex--;
    BCB_ComboBox->ItemIndex--;
    }
else
    ShowMessage("No previous items");
DirectX_ComboBoxChange(Sender);
}
//---------------------------------------------------------------------------
void __fastcall TStructures_Editor::BitBtn7Click(TObject *Sender)
{
if (DirectX_ComboBox->ItemIndex < DirectX_ComboBox->Items->Count-1)
    {
    DirectX_ComboBox->ItemIndex++;
    BCB_ComboBox->ItemIndex++;
    }
else
    ShowMessage("No more items");
DirectX_ComboBoxChange(Sender);
}
//---------------------------------------------------------------------------
void __fastcall TStructures_Editor::BitBtn4Click(TObject *Sender)
{
if (DirectX_StructureMember_ComboBox->ItemIndex > 0)
    {
    DirectX_StructureMember_ComboBox->ItemIndex--;
    BCB_ObjectProperty_ComboBox->ItemIndex--;
    }
else
    ShowMessage("No previous items");
DirectX_StructureMember_ComboBoxChange(Sender);
}
//---------------------------------------------------------------------------
void __fastcall TStructures_Editor::BitBtn5Click(TObject *Sender)
{
if (DirectX_StructureMember_ComboBox->ItemIndex < DirectX_StructureMember_ComboBox->Items->Count-1)
    {
    DirectX_StructureMember_ComboBox->ItemIndex++;
    BCB_ObjectProperty_ComboBox->ItemIndex++;
    }
else
    ShowMessage("No more items");
DirectX_StructureMember_ComboBoxChange(Sender);
}
//---------------------------------------------------------------------------
void __fastcall TStructures_Editor::Prev_ButtonClick(TObject *Sender)
{
if (DirectX_MemberFlags_ComboBox->ItemIndex > 0)
    {
    DirectX_MemberFlags_ComboBox->ItemIndex--;
    BCB_PropertyFlags_ComboBox->ItemIndex--;
    }
else
    ShowMessage("No previous items");
DirectX_MemberFlags_ComboBoxChange(Sender);
}
//---------------------------------------------------------------------------
void __fastcall TStructures_Editor::Next_ButtonClick(TObject *Sender)
{
if (DirectX_MemberFlags_ComboBox->ItemIndex < DirectX_MemberFlags_ComboBox->Items->Count-1)
    {
    DirectX_MemberFlags_ComboBox->ItemIndex++;
    BCB_PropertyFlags_ComboBox->ItemIndex++;
    }
else
    ShowMessage("No more items");
DirectX_MemberFlags_ComboBoxChange(Sender);
}
//---------------------------------------------------------------------------
void __fastcall TStructures_Editor::BCB_ComboBoxChange(TObject *Sender)
{
DirectX_ComboBox->ItemIndex = BCB_ComboBox->ItemIndex;
DirectX_ComboBoxChange(Sender);    
}
//---------------------------------------------------------------------------
void __fastcall TStructures_Editor::BCB_ObjectProperty_ComboBoxChange(
    TObject *Sender)
{
DirectX_StructureMember_ComboBox->ItemIndex = BCB_ObjectProperty_ComboBox->ItemIndex;
DirectX_StructureMember_ComboBoxChange(Sender);
}
//---------------------------------------------------------------------------
void __fastcall TStructures_Editor::BCB_PropertyFlags_ComboBoxChange(
    TObject *Sender)
{
DirectX_MemberFlags_ComboBox->ItemIndex = BCB_PropertyFlags_ComboBox->ItemIndex;
DirectX_MemberFlags_ComboBoxChange(Sender);
}
//---------------------------------------------------------------------------
void __fastcall TStructures_Editor::BCB_ObjectName_EditExit(TObject *Sender)
{
if (Structures_Table->Modified)
    {
    Structures_Table->Post();
    BuildStructureSelections();
    }
}
//---------------------------------------------------------------------------
void __fastcall TStructures_Editor::BCB_PropertyName_EditExit(
    TObject *Sender)
{
if (StructureMembers_Table->Modified)
    {
    StructureMembers_Table->Post();
    BuildMemberSelections();
    }
}
//---------------------------------------------------------------------------
void __fastcall TStructures_Editor::BCB_PropertyFlagsDescript_MemoExit(
    TObject *Sender)
{
if (StructureMemberFlags_Table->Modified)
    {
    StructureMemberFlags_Table->Post();
    BuildFlagSelections();
    }
}
//---------------------------------------------------------------------------
void __fastcall TStructures_Editor::DBRadioGroup1Exit(TObject *Sender)
{
if (Structures_Table->Modified) Structures_Table->Post();
}
//---------------------------------------------------------------------------
